//--------------------------------------------------------------------------
// Copyright (C) 1992,1993,1994,1995,1996,1997,1998 Dallas Semiconductor Corporation. 
// All rights Reserved. Printed in U.S.A.
// This software is protected by copyright laws of
// the United States and of foreign countries.
// This material may also be protected by patent laws of the United States 
// and of foreign countries.
// This software is furnished under a license agreement and/or a
// nondisclosure agreement and may only be used or copied in accordance
// with the terms of those agreements.
// The mere transfer of this software does not imply any licenses
// of trade secrets, proprietary technology, copyrights, patents,
// trademarks, maskwork rights, or any other form of intellectual
// property whatsoever. Dallas Semiconductor retains all ownership rights.
//--------------------------------------------------------------------------
//
//  NetTest.C - Test program for a network with DS2409's on the trunk
//              and DS2407's on the branches with optional DS2401's
//              connected to the A channel of the DS2407's.  Also report
//              the activity latches. 
//
//  Compiler:  Microsoft Visual C 5.0
//  Version: 1.00
//
//

#define TMEXUTIL

#include "iBTMEXCW.H"   
#include <time.h>

// constants 
// family codes by type 
#define DS2409_FAMILY   0x1F
#define DS2407_FAMILY   0x12
#define DS2401_FAMILY   0x01

// DS2407 channel states
#define AOFF_BON        0x33
#define AOFF_BOFF       0x73
#define AON_BOFF        0x53

// max number of devices at each level
#define MAX_NUM_DS2409   10
#define MAX_NUM_DS2407   20
#define MAX_NUM_DS2401   2

// DS2409 commands
#define DIRECT_ON_MAIN   0xA5
#define ALL_LINES_OFF    0x66

// Local Function Prototypes 
void main(short argc, char **argv);
short DiscoverNetwork(void);
short FindAllFamily(short Family, short SerialNum[][8], short MaxSerialNum);
short CheckNetworkActivity(void);
short SendDS2409Command(short *SerialNum, short Cmd, short NumExtra);
short SetDS2407FlipFlop(short *SerialNum, short ffstate);
short SendSkipAllOff(void);
short SendSkipAllFlipFlopOpen(void);
void DisplayNetwork(void);
short ReadSwitchInfo(short *SerialNum, short ClearActivity, short *Info);
short SeviceDS2407(short HSNumber);
void output(short print_flag, char *msg);
void Exit_Prog(char *, short);
void PrintSerialNum(short *SerialNumber);
ushort crc16(int data);
uchar dowcrc(uchar);

// Global Variables to hold list information 
short PortType, PortNum;
uchar state_buffer[5120];
short WeatherROM[3][8];
long session_handle;                         
ushort CRC16;
uchar CRC8;                   
short Verbose=0;
FILE *fp;

// ROM numbers
short HighSwitch[MAX_NUM_DS2409][8], HSCnt=0;
short LowSwitch[MAX_NUM_DS2409][MAX_NUM_DS2407][8], LSCnt[MAX_NUM_DS2409];
short ID[MAX_NUM_DS2409][MAX_NUM_DS2407][MAX_NUM_DS2401][8], IDCnt[MAX_NUM_DS2409][MAX_NUM_DS2407];


//----------------------------------------------------------------------
//  This is the Main routine for NetTest
// 
void main(short argc, char **argv)
{
   short printmsg=1,found,ClearActivity=0,filenum;
   short foundnetwork=0,didsetup=0,rslt,errorcnt=0;
   char tstr[80];       
   char ReturnString[256];
   HKEY GlobKey;
   DWORD dwType,size;
   char temp_zstr[256],ch;

   // check arguments to see if request instruction with '?' or too many 
   if ((argc > 3) || ((argc > 1) && (argv[1][0] == '?' || argv[1][1] == '?')))
   {
       printf("\nusage: TestNet <output_filename>\n"
              "  - Test program for a network with DS2409's on the trunk\n"
              "    and DS2407's on the branches with optional DS2401's\n"
              "    connected to the A channel of the DS2407's.\n"
              "  - Uses the default MicroLAN port (use SetPort32.exe to set)\n"
              "  - <output_filename> optional output filename\n"
              "  - press any key to stop test\n"
              "  - version 1.00TEST1\n");
       Exit_Prog("",BAD_ARGUMENT);
   }

   // read the default PortNum and PortType from the registry 
   if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
         "Software\\Dallas Semiconductor\\iButton TMEX\\3.00",0,
         KEY_READ,&GlobKey) == ERROR_SUCCESS)
   {
      // attempt to read the PortNum 
      size = 255;
      PortNum = 1;
      if (RegQueryValueEx(GlobKey,"DefaultPortNum",NULL,&dwType,
          &ReturnString[0],&size) == ERROR_SUCCESS)
      {
         if ((dwType == REG_SZ) || (size > 0))
            PortNum = atoi(ReturnString);
      }
      // attempt to read the PortType 
      size = 255;
      PortType = 1;
      if (RegQueryValueEx(GlobKey,"DefaultPortType",NULL,&dwType,
          &ReturnString[0],&size) == ERROR_SUCCESS)
      {
         if ((dwType == REG_SZ) || (size > 0))
            PortType = atoi(ReturnString);
      }

      // close the key
      RegCloseKey(GlobKey);
   }
   else
     Exit_Prog("ERROR, Could not read the default PortNum and PortType from registry!\n",
               NO_DRIVERS);

   // check to see that indicated PortNum is  <= 15 
   if ((PortNum > 15) || (PortType > 15))
     Exit_Prog("ERROR, Indicated PortNum is not valid!\n",
               -XONE_WIRE_PORT_ERROR);

   // check arguments 
   Verbose = FALSE;
   filenum = 0;
   if (argc >= 2)
   {
      if (argv[1][0] != '/')
         filenum = 1;
      else if ((argv[1][1] == 'V') || (argv[1][1] == 'v'))
         Verbose = TRUE;

      if (argc == 3)
      {    
         if (argv[2][0] != '/')
            filenum = 2;
         else if ((argv[2][1] == 'V') || (argv[2][1] == 'v'))
            Verbose = TRUE;
      }   
   }

   // open the output file  
   fp = NULL;
   if (filenum > 0)
   {
      fp = fopen(argv[filenum],"w+");
      if(fp == NULL)
         Exit_Prog("ERROR, Could not open output file!\n",0);
   }

   // print a header   
   output(TRUE,"\nTestNet Version 1.00TEST1\n"
         "Dallas Semiconductor Corporation\n"
         "Copyright (C) 1998\n\n");
   sprintf(temp_zstr,"Default Port number: %d    Port type: %d\n",PortNum,PortType);
   output(TRUE,temp_zstr);   
   Get_Version(tstr);
   sprintf(temp_zstr,"TMEX main driver: %s\n",tstr);
   output(TRUE,temp_zstr);   
   sprintf(temp_zstr,"TMEX adapter driver #%d:",PortType);
   output(TRUE,temp_zstr);   
   if (TMGetTypeVersion(PortType,tstr) < 0)
      Exit_Prog("No Hardware Driver for this type found!",NO_DRIVERS);
   sprintf(temp_zstr," %s\n\n",tstr);
   output(TRUE,temp_zstr);   
   output(TRUE,"\nPress N to re-discover the entire netork or X to quit.\n\n");

   // loop to get a session handle, find weather station, read 
   found = FALSE;
   do
   {
      // attempt to get a session 
      session_handle = TMExtendedStartSession(PortNum,PortType,NULL);
      if (session_handle > 0)  
      {
         // reset the session prompt 
         printmsg = TRUE;
 
         // setup the MicroLan 
         if (!didsetup)
         {
            rslt = TMSetup(session_handle);
            sprintf(temp_zstr,"Result of setup of the MicroLAN: %d\n\n",rslt);
            output(TRUE,temp_zstr);
            // check for good setup 
            if (rslt == 1)
               didsetup = TRUE;
            else
               break;
         }

         // find the Weather Station 
         if (didsetup)
         {
            // check for a network, require at least 1 DS2409 and 1 DS2407 
            if (!foundnetwork)
            {
               if (DiscoverNetwork())
               {
                  foundnetwork = TRUE;
                  DisplayNetwork();
               }
               else
               {
                  output(TRUE, "\nError reading the network\n");
                  break;
               }
            }

            // check if any DS2407's show activity 
            if (CheckNetworkActivity())
               errorcnt = 0;   // reset the error count 
            else
            {            
               errorcnt++;
               sprintf(temp_zstr,"Failed to read activity.\n",errorcnt);
               output(TRUE, temp_zstr);
               // check for abort 
               if (errorcnt > 4)
                  break;
            }

         }   
      }
      else if (session_handle == 0)
      {                            
         // check if need to print the waiting message 
         if (printmsg)
         {  
            printmsg = FALSE;
            output(TRUE,"\nWaiting to get access to the MicroLan\n\n");
         }
      }
      else if (session_handle < 0)   
      {
         MessageBeep(MB_ICONEXCLAMATION);
         output(TRUE,"No TMEX adapter driver for this type found!\n");
         break;
      }

      // end the session if one is open 
      TMEndSession(session_handle);

      // be friendly to other applications,sleep for 30ms 
      Sleep(30);

      // check for a key hit
      if (kbhit())
      {
         ch = getch();
         printf("Key pressed %02X (X to quit)\n",ch);
      }

      // check for re-read of network
      if ((ch == 'n') || (ch == 'N'))
      {
         foundnetwork = FALSE;
         ch = 0;
      }
   }
   while ((ch != 'X') && (ch != 'x'));

   // end the session if one is open 
   TMEndSession(session_handle);
}


//----------------------------------------------------------------------
// Discover the current network                                                           
//
short DiscoverNetwork(void)
{    
   short h,l;
   char temp_zstr[256];
   ulong start_time, end_time;   

   // get a time stamp (mS)
   start_time = GetTickCount();         

   // send skip and 'all-off' command (open all DS2409's)
   SendSkipAllOff();

   // read all of the 
   HSCnt = FindAllFamily(DS2409_FAMILY, HighSwitch, MAX_NUM_DS2409);

   // loop for each DS2409, open, search for DS2407, close
   for (h = 0; h < HSCnt; h++)
   {
      // close DS2409 main channel
      if (SendDS2409Command(HighSwitch[h],DIRECT_ON_MAIN,0))
      {
         // send skip and all close DS2407 command 
         SendSkipAllFlipFlopOpen();

         // find all DS2407 on this branch
         LSCnt[h] = FindAllFamily(DS2407_FAMILY, LowSwitch[h], MAX_NUM_DS2407);

         // loop through each DS2407 found and look for DS2401's on A channel
         for (l = 0; l < LSCnt[h]; l++)
         {
            // close channel A of DS2407
            if (SetDS2407FlipFlop(LowSwitch[h][l], AON_BOFF))
            {
               // find all DS2401's       
               IDCnt[h][l] = FindAllFamily(DS2401_FAMILY, ID[h][l], MAX_NUM_DS2401);

               // open the channel A of the DS2407
               if (!SetDS2407FlipFlop(LowSwitch[h][l], AOFF_BOFF))
                  return FALSE;
            }               
         }

         // open the DS2409 main channel
         if (!SendDS2409Command(HighSwitch[h],ALL_LINES_OFF,0))
            return FALSE;
      }
      else
         return FALSE;
   }

   // get a time stamp (mS)
   end_time = GetTickCount();         

   // print time
   sprintf(temp_zstr,"\nSuccessful discovery of network in %dmS\n",end_time-start_time);
   output(TRUE,temp_zstr);

   return TRUE;
}

//----------------------------------------------------------------------
// Print the current state of the network
//
void DisplayNetwork(void)
{
   int i,h,l,cnt=0;
   char zstr_tmp[255];

   output(TRUE,"\nPrint discovered network:\n\n");

   // loop to display the current network
   for (h = 0; h < HSCnt; h++)
   {
      sprintf(zstr_tmp,"(%d) ",h);
      output(TRUE,zstr_tmp);
      PrintSerialNum(HighSwitch[h]);
      output(TRUE,"\n");

      for (l = 0; l < LSCnt[h]; l++)
      {
         sprintf(zstr_tmp,"       (%d) ",l);
         output(TRUE,zstr_tmp);
         PrintSerialNum(LowSwitch[h][l]);
         output(TRUE,"\n");

         for (i = 0; i < IDCnt[h][l]; i++)
         {
            sprintf(zstr_tmp,"              (%d) ",i);
            output(TRUE,zstr_tmp);
            PrintSerialNum(ID[h][l][i]);
            output(TRUE,"\n");
         }
      }
   }
   output(TRUE,"\n");
}


//----------------------------------------------------------------------
// Use the activity latches to detect new DS2401's or a switch
//
short CheckNetworkActivity(void)
{
   short h,rslt,activity=0;
   ulong start_time, end_time;   
   char zstr_tmp[255];

   // get a time stamp (mS)
   start_time = GetTickCount();         

   // send skip and 'all-off' command (open all DS2409's)
   SendSkipAllOff();

   // loop for each DS2409, open, search for alarming DS2407, close
   for (h = 0; h < HSCnt; h++)
   {
      // close DS2409 main channel
      if (SendDS2409Command(HighSwitch[h],DIRECT_ON_MAIN,0))
      {
         rslt = TMFirstAlarm(session_handle,state_buffer);

         // loop while alarming devices (DS2407) detected
         while (rslt == 1)
         {
            // set flag indicating activity
            activity = TRUE;

            // read it status
            if (!SeviceDS2407(h))
               return FALSE;

            // look for next
            rslt = TMNextAlarm(session_handle,state_buffer);
         }

         // open the DS2409 main channel
         if (!SendDS2409Command(HighSwitch[h],ALL_LINES_OFF,0))
            return FALSE;
      }

   }

   // get a time stamp (mS)
   end_time = GetTickCount();         

   // print time if activity
   if (activity)
   {
      sprintf(zstr_tmp,"   Activity check completed in %dmS\n",end_time-start_time);
      output(TRUE,zstr_tmp);
   }

   return TRUE; 
}


//----------------------------------------------------------------------
// Service the current DS2407 which has been found to be
// alarming.  Read/reset the activity latches and re-discover
// the DS2401's connected to channel A.  
// 
short SeviceDS2407(short HSNumber)
{
   short SerialNum[8],Info,l,i;
   char zstr_tmp[255];

   // read the rom number of the device
   SerialNum[0] = 0;
   TMRom(session_handle,state_buffer,SerialNum);   

   // loop throught the DS2407's on branch 'HSNumber'
   // to find out which one it is
   for (l = 0; l < LSCnt[HSNumber]; l++)
   {
      for (i = 0; i < 8; i++)
         if (LowSwitch[HSNumber][l][i] != SerialNum[i])
            break;

      // match found
      if (i == 8)
         break;
   }

   // check for no match
   if (l == LSCnt[HSNumber])
      return FALSE;

   // print the output
   output(TRUE,"\nActivity: branch ");
   PrintSerialNum(HighSwitch[HSNumber]);
   output(TRUE," checking DS2407 ");
   PrintSerialNum(LowSwitch[HSNumber][l]);
   
   // read the status
   if (ReadSwitchInfo(LowSwitch[HSNumber][l],TRUE,&Info))
   {
      sprintf(zstr_tmp," status %02X\n",Info);
      output(TRUE,zstr_tmp);
   }
   else
      return FALSE;
   
   // update the devices connect to channel A
   // close channel A of DS2407
   if (SetDS2407FlipFlop(LowSwitch[HSNumber][l], AON_BOFF))
   {
      // find all DS2401's       
      IDCnt[HSNumber][l] = FindAllFamily(DS2401_FAMILY, ID[HSNumber][l], MAX_NUM_DS2401);

      // open the channel A of the DS2407
      if (!SetDS2407FlipFlop(LowSwitch[HSNumber][l], AOFF_BOFF))
         return FALSE;
   }               

   // print the current
   output(TRUE,"   Devices connected to channel A\n");
   for (i = 0; i < IDCnt[HSNumber][l]; i++)
   {
      sprintf(zstr_tmp,"              (%d) ",i);
      output(TRUE,zstr_tmp);
      PrintSerialNum(ID[HSNumber][l][i]);
      output(TRUE,"\n");
   }

   // put the serial number back to continue search
   TMRom(session_handle,state_buffer,SerialNum);   

   return TRUE;
}


//----------------------------------------------------------------------
// Broadcast the 'all lines off' command to the DS2409's to
// disable any left on. 
// 
short SendSkipAllOff(void)
{
   uchar send_block[50];
   short send_cnt,repeat=0;

   send_cnt = 0; 
   // add the skip command
   send_block[send_cnt++] = 0xCC;

   // send the all lines off command
   send_block[send_cnt++] = ALL_LINES_OFF;

   // read the confirmation
   send_block[send_cnt++] = 0xFF;

   // send the command string 
   if (TMBlockIO(session_handle,send_block,send_cnt) == send_cnt)
      return TRUE;
   else
      return FALSE;
}

//----------------------------------------------------------------------
// Broadcast the 'all channels open' command to the DS2407's to
// disable any left closed. 
//
short SendSkipAllFlipFlopOpen(void)
{
   uchar send_block[50];
   short send_cnt,repeat=0;

   send_cnt = 0; 
   // add the skip command
   send_block[send_cnt++] = 0xCC;

   // write status command
   send_block[send_cnt++] = 0x55;
   // address of switch state
   send_block[send_cnt++] = 0x07;
   send_block[send_cnt++] = 0x00;
   // write state
   send_block[send_cnt++] = AOFF_BOFF;
   // read CRC16
   send_block[send_cnt++] = 0xFF;
   send_block[send_cnt++] = 0xFF;

   // send the command string 
   if (TMBlockIO(session_handle,send_block,send_cnt) == send_cnt)
      return TRUE;
   else
      return FALSE;
}


//----------------------------------------------------------------------
// Send and comfirm a command sent to a DS2409.  If it fails try up 
// to 10 times.
//
short SendDS2409Command(short *SerialNum, short Cmd, short NumExtra)
{
   uchar send_block[50];
   short send_cnt,i,rslt,repeat=0;

   do
   {
      send_cnt = 0; 
      // add the match command
      send_block[send_cnt++] = 0x55;
      for (i = 0; i < 8; i++)
         send_block[send_cnt++] = (uchar)SerialNum[i];

      // the command
      send_block[send_cnt++] = (uchar)Cmd;

      // extra bytes and confirmation
      for (i = 0; i <= NumExtra; i++)
         send_block[send_cnt++] = 0xFF;

      // send the command string 
      rslt = TMBlockIO(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
        if (send_block[send_cnt - 1] == Cmd) 
            return TRUE;
      }

      // delay to let things settle
      Sleep(10);
   }
   while (repeat++ < 10);

   return FALSE;       
}


//----------------------------------------------------------------------
// Find all of the devices of a particular 'Family' type, return the
// number of devices found.
//
short FindAllFamily(short Family, short SerialNum[][8], short MaxSerialNum)
{ 
   short sn_cnt = 0;

   // search for the DS2409's on the trunk
   if (TMFamilySearchSetup(session_handle,&state_buffer,Family) == 1)
   { 
      // loop to get all of the DS2409's
      for (;;)
      {
         if (TMNext(session_handle,state_buffer) == 1)
         {
            // read the rom number 
            SerialNum[sn_cnt][0] = 0;
            TMRom(session_handle,state_buffer,&SerialNum[sn_cnt][0]);   

            // check if correct type 
            if ((SerialNum[sn_cnt][0] & 0x7F) != Family)
               break;
            
            // increment and check for too many 
            if (++sn_cnt >= MaxSerialNum)
               break;
         }
         else
            break;
      }
   }
   
   return sn_cnt;
}


//----------------------------------------------------------------------
// Set the DS2407 flip flops to the 'ffstate' .
//                                                          
short SetDS2407FlipFlop(short *SerialNum, short ffstate)
{    
   short rt=FALSE,i,rslt,repeat=0;
   uchar send_block[30];
   short send_cnt=0;

   do
   {
      // create a block to send the sets the state
      send_cnt = 0; 
      CRC16 = 0;
      // add the match command
      send_block[send_cnt++] = 0x55;
      for (i = 0; i < 8; i++)
         send_block[send_cnt++] = (uchar)SerialNum[i];

      // write status command
      send_block[send_cnt++] = 0x55;
      crc16(0x55);
      // address of switch state
      send_block[send_cnt++] = 0x07;
      crc16(0x07);
      send_block[send_cnt++] = 0x00;
      crc16(0x00);
      // write state
      send_block[send_cnt++] = (uchar)ffstate;
      crc16(ffstate);
      // read CRC16
      send_block[send_cnt++] = 0xFF;
      send_block[send_cnt++] = 0xFF;

      // now send the block
      rslt = TMBlockIO(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
         // perform the CRC16 on the last 2 bytes of packet
         crc16(send_block[send_cnt-2]);
         crc16(send_block[send_cnt-1]);

         // verify CRC16 is correct
         if (CRC16 == 0xB001)
            // success
            return TRUE;
      }

      // delay to let things settle
      Sleep(10);
   }
   while (repeat++ < 10);

   return FALSE;       
}


//----------------------------------------------------------------------
// Read the switch information byte.
//                                                           
short ReadSwitchInfo(short *SerialNum, short ClearActivity, short *Info)
{
   short rt=FALSE,i,rslt,repeat=0;
   uchar send_block[30];
   short send_cnt=0;

   do
   {
      // create a block to send the sets the state
      send_cnt = 0; 
      CRC16 = 0;
      // add the match command
      send_block[send_cnt++] = 0x55;
      for (i = 0; i < 8; i++)
         send_block[send_cnt++] = (uchar)SerialNum[i];

      // channel access command 
      send_block[send_cnt++] = 0xF5;
      crc16(0xF5);

      // control bytes                  
      if (ClearActivity)
      {
         send_block[send_cnt++] = 0xD5;
         crc16(0xD5);
      }
      else
      {
         send_block[send_cnt++] = 0x55;
         crc16(0x55);
      }
      send_block[send_cnt++] = 0xFF;
      crc16(0xFF);
   
      // read the info byte, dummy and CRC16
      for (i = 0; i < 4; i++)
         send_block[send_cnt++] = 0xFF;

      // now send the block
      rslt = TMBlockIO(session_handle,send_block,send_cnt);
      if (rslt == send_cnt)
      {
         // calculate the CRC16 on the CRC16 bytes to get B001
         for (i = send_cnt - 4; i < send_cnt; i++)
            crc16(send_block[i]);

         // get the info value
         *Info = send_block[send_cnt - 4];
   
         // check result 
         if (CRC16 == 0xB001) 
            return TRUE;
      }

      // delay to let things settle
      Sleep(10);
   }
   while (repeat++ < 10);


   return FALSE;
}


//--------------------------------------------------------------------------
// Update the Dallas Semiconductor One Wire CRC (CRC8) from the global
// variable CRC8 and the argument.  Return the updated CRC8.
// 
uchar dscrc_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53};

uchar dowcrc(uchar x)
{
   CRC8 = dscrc_table[CRC8 ^ x];
   return CRC8;
}


 
//--------------------------------------------------------------------------
//  Prints a message, beeps and exits the program.
// 
void Exit_Prog(char *msg, short errcd)
{
   // print the message left justified 
   printf("\r%s",msg);
   
   //  error beep if return code is not 0 
   if (errcd != NORMAL_EXIT)
     MessageBeep(MB_ICONEXCLAMATION);
   
   //  return error code passed 
   exit(errcd);
}


//--------------------------------------------------------------------------
//  Conditionally prints a message, 
// 
void output(short print_flag, char *msg)
{
   // only print following in verbose mode 
   if (print_flag)
   {
      printf("%s",msg);
      if (fp != NULL)
         fprintf(fp,"%s",msg);
   }
}

//----------------------------------------------------------------------
// Read and print the Serial Number.
//
void PrintSerialNum(short *SerialNumber)
{
   int i,cnt=0;
   char zstr_tmp[50];

   for (i = 7; i >= 0; i--)
      cnt += sprintf(&zstr_tmp[cnt],"%02X",SerialNumber[i]);

   output(TRUE,zstr_tmp);
}

//--------------------------------------------------------------------------
// Calculate a new CRC16 from the input data integer.  Return the current
// CRC16 and also update the global variable CRC16.
//
// 'data' - data to perform a CRC16 on
//
// Returns: the current CRC16
//
static short oddparity[16] = { 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0 };

ushort crc16(int data)
{
   data = (data ^ (CRC16 & 0xff)) & 0xff;
   CRC16 >>= 8;

   if (oddparity[data & 0xf] ^ oddparity[data >> 4])
      CRC16 ^= 0xc001;

   data <<= 6;
   CRC16   ^= data;
   data <<= 1;
   CRC16   ^= data;

   return CRC16;
}

